# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
CDLExecEnv test
---------------

Checking DataLab execution environment management.
"""

# guitest: skip (only needed for regression tests)

from __future__ import annotations

import os
import sys

import cdl
from cdl.env import VerbosityLevels, execenv
from cdl.utils.tests import get_script_output

ARGV_TEST = "--execenvtest"


def print_execenv() -> None:
    """Print CDL execution environment"""
    sys.argv.remove(ARGV_TEST)
    print(str(execenv.to_dict()))


def get_subprocess_execenv_dict(args: list[str], env: dict | None = None) -> dict:
    """Get CDL execution environment dict from subprocess

    Args:
        args (list[str]): command-line arguments
        env (dict | None): environment variables to pass to subprocess

    Returns:
        dict[str, str | int | bool | None]: CDL execution environment dict
    """
    output = get_script_output(__file__, args=args + [ARGV_TEST], env=env)
    return eval(output)  # pylint: disable=eval-used


def assert_two_dicts_are_equal(
    dict1: dict, dict2: dict, exceptions: tuple[str] | None = None
) -> None:
    """Assert two dicts are equal

    Args:
        dict1 (dict): first dict
        dict2 (dict): second dict
        exceptions (tuple[str] | None): keys to ignore
    """
    diff_keys = []
    for key in dict1:
        if key in exceptions:
            continue
        if dict1[key] != dict2[key]:
            diff_keys.append(key)
    if diff_keys:
        assert False, f"Dictionaries differ on keys: {diff_keys}"


def test_cli():
    """Test CDL execution environment from command-line"""
    remove_all_cdl_envvars()

    # Test default values
    execenv.print("Testing command-line arguments:")
    execenv.print("  Default values: ", end="")
    execenvdict = get_subprocess_execenv_dict([])
    execenv.print("OK")
    assert execenvdict == execenv.to_dict()
    # Testing boolean arguments
    execenv.print("  Testing boolean arguments:")
    for argname in ("unattended", "accept_dialogs", "screenshot"):
        execenv.print(f"    {argname}:", end="")
        for val in (True, False):
            execenv.print(f" {val}", end="")
            if val:
                args = [f"--{argname}"]
            else:
                args = []  # Default value is False
            execenvdict = get_subprocess_execenv_dict(args)
            assert_two_dicts_are_equal(execenvdict, execenv.to_dict(), (argname,))
            assert (
                execenvdict[argname] is val
            ), f"execenvdict[{argname}] = {execenvdict[argname]} != {val}"
        execenv.print()
    # Testing integer arguments
    execenv.print("  Testing integer arguments:")
    for argname in ("delay", "xmlrpcport"):
        execenv.print(f"    {argname}:", end="")
        for val in (None, 0, 1, 2):
            if val is None:
                args = []
            else:
                args = [f"--{argname}", str(val)]
            execenv.print(f" {val}", end="")
            execenvdict = get_subprocess_execenv_dict(args)
            assert_two_dicts_are_equal(execenvdict, execenv.to_dict(), (argname,))
            if val is None:
                if argname == "delay":
                    # Default value is 0 for delay
                    assert execenvdict[argname] == 0
                else:
                    # Default value is None for xmlrpcport
                    assert execenvdict[argname] is None
            else:
                assert execenvdict[argname] == val
        execenv.print()
    # Testing choice arguments
    execenv.print("  Testing choice arguments:")
    for argname in ("verbose",):
        execenv.print(f"    {argname}:", end="")
        choices = {"verbose": [verb.value for verb in VerbosityLevels]}
        defaultval = {"verbose": VerbosityLevels.NORMAL.value}[argname]
        for val in [None] + choices[argname]:
            if val is None:
                args = []
            else:
                args = [f"--{argname}", val]
            execenv.print(f" {val}", end="")
            execenvdict = get_subprocess_execenv_dict(args)
            assert_two_dicts_are_equal(execenvdict, execenv.to_dict(), (argname,))
            if val is None:
                assert execenvdict[argname] == defaultval
            else:
                assert execenvdict[argname] == val
        execenv.print()
    # Testing special arguments
    execenv.print("  Testing special arguments:")
    execenv.print("    version: ", end="")
    args = ["--version"]
    output = get_script_output(__file__, args=args + [ARGV_TEST])
    execenv.print(output)
    assert cdl.__version__ in output
    execenv.print("    h5 positionnal argument: ", end="")
    for h5files in (None, ["test.h5"], ["toto.h5", "tata.h5"]):
        if h5files is None:
            args = []
        else:
            args = [";".join(h5files)]
        execenv.print(f" {h5files}", end="")
        execenvdict = get_subprocess_execenv_dict(args)
        assert execenvdict["h5files"] == h5files
    execenv.print()
    execenv.print("    h5browser argument:")
    for argname in ("-b", "--h5browser"):
        execenv.print(f"      {argname}", end="")
        for val in (None, "test.h5"):
            if val is None:
                args = []
            else:
                args = [argname, val]
            execenv.print(f" {val}", end="")
            execenvdict = get_subprocess_execenv_dict(args)
            assert execenvdict["h5browser_file"] == val
        execenv.print()
    execenv.print("=> Everything is OK")


def remove_all_cdl_envvars():
    """Remove all CDL environment variables"""
    for _attrname, envvar in execenv.iterate_over_attrs_envvars():
        os.environ.pop(envvar, None)


def get_attr_to_envvar(
    vartype: type,
    default: int | str | None | bool,
    values: list[int | str | None | bool] | None = None,
) -> tuple:
    """Get ATTR_TO_ENVVAR tuple for a given type"""
    if vartype is bool:
        if default is False:
            ate = ((True, ("1", "true")), (False, (None, "", "0", "false")))
        else:
            ate = ((True, (None, "", "1", "true")), (False, ("0", "false")))
    elif vartype is int:
        ate = []
        for val in values:
            if val == default:
                ate.append((val, (None, "", str(val))))
            else:
                ate.append((val, (str(val),)))
        ate = tuple(ate)
        if default is None:
            ate = ((None, (None, "")),) + ate
    elif vartype is str:
        ate = ((None, (None, "")),)
    elif vartype is list:
        ate = []
        for choice in values:
            if choice == default:
                ate.append((choice, (None, "", choice)))
            else:
                ate.append((choice, (choice,)))
        ate = tuple(ate)
    else:
        raise ValueError(f"Unknown type {vartype}")
    return ate


ATTR_TO_ENVVAR = {
    "unattended": get_attr_to_envvar(bool, False),
    "accept_dialogs": get_attr_to_envvar(bool, False),
    "screenshot": get_attr_to_envvar(bool, False),
    "do_not_quit": get_attr_to_envvar(bool, False),
    "catcher_test": get_attr_to_envvar(bool, False),
    "delay": get_attr_to_envvar(int, 0, [0, 10, 20]),
    "xmlrpcport": get_attr_to_envvar(int, None, [9854, 1020, 213]),
    "verbose": get_attr_to_envvar(
        list, VerbosityLevels.NORMAL.value, [verb.value for verb in VerbosityLevels]
    ),
}


def test_envvar():
    """Testing DataLab configuration file"""
    assert execenv.unattended is False, "This test must be run with unattended=False"
    print("Testing DataLab execution environment:")
    for attrname, envvar in execenv.iterate_over_attrs_envvars():
        attr_to_envvar = ATTR_TO_ENVVAR[attrname]
        print(f"  Testing {attrname}:")
        for value, envvals in attr_to_envvar:
            print(f"    {value}: [attr->env]", end="")
            remove_all_cdl_envvars()
            if value is not None:
                setattr(execenv, attrname, value)
            assert (
                os.environ.get(envvar) in envvals
            ), f"os.environ[{envvar}] = {os.environ.get(envvar)} != {envvals}"
            print(" [env->attr]", end="")
            remove_all_cdl_envvars()
            for envval in envvals:
                if envval is not None:
                    os.environ[envvar] = envval
                assert getattr(execenv, attrname) == value, (
                    f"execenv.{attrname} = {getattr(execenv, attrname)}"
                    f" != {value} (envval = {repr(envval)})"
                )
            print(" [env->subprocess->attr]", end="")
            remove_all_cdl_envvars()
            for envval in envvals:
                if envval is not None:
                    os.environ[envvar] = envval
                execenvdict = get_subprocess_execenv_dict([])
                assert execenvdict[attrname] == value, (
                    f"execenvdict[{attrname}] = {execenvdict[attrname]}"
                    f" != {value} (envval = {repr(envval)})"
                )
            print()
    print("=> Everything is OK")


if __name__ == "__main__":
    if ARGV_TEST in sys.argv:
        print_execenv()
    else:
        test_envvar()
        test_cli()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
GUI
===

The :mod:`cdl.core.gui` package contains functionnalities related to the graphical
user interface (GUI) of the DataLab (CDL) project. Those features are mostly specific
to DataLab and are not intended to be used independently.

The purpose of this section of the documentation is to provide an overview of the
DataLab GUI architecture and to describe the main features of the modules contained
in this package. It is not intended to provide a detailed description of the GUI
features, but rather to provide a starting point for the reader who wants to
understand the DataLab internal architecture.

DataLab's main window is composed of several parts, each of them being handled by a
specific module of this package:

- **Signal and image panels**: those panels are used to display signals and images
  and to provide a set of tools to manipulate them. Each data panel relies on a set
  of modules to handle the GUI features (:mod:`cdl.core.gui.actionhandler` and
  :mod:`cdl.core.gui.objectview`), the data model (:mod:`cdl.core.gui.objectmodel`),
  the data visualization (:mod:`cdl.core.gui.plothandler`),
  and the data processing (:mod:`cdl.core.gui.processor`).

- **Macro panel**: this panel is used to display and run macros. It relies on the
  :mod:`cdl.core.gui.macroeditor` module to handle the macro edition and execution.

- **Specialized widgets**: those widgets are used to handle specific features such as
  ROI edition (:mod:`cdl.core.gui.roieditor`),
  Intensity profile edition (:mod:`cdl.core.gui.profiledialog`), etc.

.. list-table::
    :header-rows: 1
    :align: left

    * - Submodule
      - Purpose

    * - :mod:`cdl.core.gui.main`
      - DataLab main window and application

    * - :mod:`cdl.core.gui.panel`
      - Signal, image and macro panels

    * - :mod:`cdl.core.gui.actionhandler`
      - Application actions (menus, toolbars, context menu)

    * - :mod:`cdl.core.gui.objectview`
      - Widgets to display object (signal/image) trees

    * - :mod:`cdl.core.gui.plothandler`
      - `PlotPy` plot items for representing signals and images

    * - :mod:`cdl.core.gui.roieditor`
      - ROI editor

    * - :mod:`cdl.core.gui.processor`
      - Processor

    * - :mod:`cdl.core.gui.docks`
      - Dock widgets

    * - :mod:`cdl.core.gui.h5io`
      - HDF5 input/output

"""

import abc

from guidata.io import BaseIOHandler


class ObjItf(abc.ABC):
    """Interface for objects handled by panels"""

    @property
    @abc.abstractmethod
    def title(self) -> str:
        """Object title"""

    @abc.abstractmethod
    def regenerate_uuid(self):
        """Regenerate UUID

        This method is used to regenerate UUID after loading the object from a file.
        This is required to avoid UUID conflicts when loading objects from file
        without clearing the workspace first.
        """

    @abc.abstractmethod
    def serialize(self, writer: BaseIOHandler):
        """Serialize this object"""

    @abc.abstractmethod
    def deserialize(self, reader: BaseIOHandler):
        """Deserialize this object"""



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Configuration test

Checking .ini configuration file management.
"""

# guitest: show

import os
import os.path as osp

from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from cdl import app
from cdl.config import Conf
from cdl.env import execenv
from cdl.utils.conf import CONF
from cdl.utils.qthelpers import cdl_app_context
from cdl.utils.tests import get_test_fnames

SEC_MAIN = Conf.main
OPT_MAX = SEC_MAIN.window_maximized
OPT_POS = SEC_MAIN.window_position
OPT_SIZ = SEC_MAIN.window_size
OPT_DIR = SEC_MAIN.base_dir

SEC_CONS = Conf.console
OPT_CON = SEC_CONS.console_enabled

CONFIGS = (
    {
        SEC_MAIN.get_name(): {
            OPT_MAX.option: False,
            OPT_POS.option: (250, 250),
            OPT_SIZ.option: (1300, 700),
            OPT_DIR.option: "",
        },
        SEC_CONS.get_name(): {
            OPT_CON.option: False,
        },
    },
    {
        SEC_MAIN.get_name(): {
            OPT_MAX.option: False,
            OPT_POS.option: (100, 100),
            OPT_SIZ.option: (750, 600),
            OPT_DIR.option: osp.dirname(__file__),
        },
        SEC_CONS.get_name(): {
            OPT_CON.option: False,
        },
    },
    {
        SEC_MAIN.get_name(): {
            OPT_MAX.option: True,
            OPT_POS.option: (10, 10),
            OPT_SIZ.option: (750, 600),
            OPT_DIR.option: "",
        },
        SEC_CONS.get_name(): {
            OPT_CON.option: True,
        },
    },
)


def apply_conf(conf, name):
    """Apply configuration"""
    execenv.print(f"  Applying configuration {name}:")
    fname = CONF.filename()
    try:
        os.remove(fname)
        execenv.print(f"    Removed configuration file {fname}")
    except FileNotFoundError:
        execenv.print(f"    Configuration file {fname} was not found")
    for section, settings in conf.items():
        for option, value in settings.items():
            execenv.print(f"    Writing [{section}][{option}] = {value}")
            CONF.set(section, option, value)
    CONF.save()


def is_wsl() -> bool:
    """Return True if running on Windows Subsystem for Linux (WSL)"""
    if os.name == "posix":
        return "WSL" in os.uname().release  # pylint: disable=no-member
    return False


def assert_in_interval(val1, val2, interval, context):
    """Raise an AssertionError if val1 is in [val2-interval/2,val2+interval/2]"""
    itv1, itv2 = val2 - 0.5 * interval, val2 + 0.5 * interval
    try:
        assert itv1 <= val1 <= itv2
    except AssertionError as exc:
        if is_wsl():
            # Ignore if executing the test on WSL: position of windows is not reliable
            # on WSL (e.g. Gnome) and the test will fail
            pass
        else:
            raise AssertionError(f"{context}: {itv1} <= {val1} <= {itv2}") from exc


def check_conf(conf, name, win: QW.QMainWindow, h5files):
    """Check configuration"""
    execenv.print(f"  Checking configuration {name}: ")
    sec_main_name = SEC_MAIN.get_name()
    sec_cons_name = SEC_CONS.get_name()
    sec_main = conf[sec_main_name]
    sec_cons = conf[sec_cons_name]
    execenv.print(f"    Checking [{sec_main_name}][{OPT_MAX.option}]: ", end="")
    assert sec_main[OPT_MAX.option] == (
        win.windowState() == QC.Qt.WindowState.WindowMaximized
    )
    execenv.print("OK")
    execenv.print(f"    Checking [{sec_main_name}][{OPT_POS.option}]: ", end="")
    if not sec_main[OPT_MAX.option]:  # Check position/size only when not maximized
        #  Check position, taking into account screen offset (e.g. Linux/Gnome)
        conf_x, conf_y = sec_main[OPT_POS.option]
        conf_w, conf_h = sec_main[OPT_SIZ.option]
        available_go = QW.QDesktopWidget().availableGeometry()
        x_offset, y_offset = available_go.x(), available_go.y()
        assert_in_interval(win.x(), conf_x - x_offset, 0, "X position")
        assert_in_interval(win.y(), conf_y - y_offset / 2, 15 + y_offset, "Y position")
        #  Check size
        assert_in_interval(win.width(), conf_w, 5, "Width")
        assert_in_interval(win.height(), conf_h, 5, "Height")
        execenv.print(f"OK {win.x(), win.y(), win.width(), win.height()}")
    else:
        execenv.print("Passed (maximized)")
    execenv.print(f"    Checking [{sec_cons_name}][{OPT_CON.option}]: ", end="")
    assert sec_cons[OPT_CON.option] == (win.console is not None)
    execenv.print("OK")
    execenv.print(f"    Checking [{sec_main_name}][{OPT_DIR.option}]: ", end="")
    if h5files is None:
        assert conf[SEC_MAIN.get_name()][OPT_DIR.option] == OPT_DIR.get()
        execenv.print("OK (written in conf file)")
    else:
        assert OPT_DIR.get() == osp.dirname(h5files[0])
        execenv.print("OK (changed to HDF5 file path)")


def test_config():
    """Testing DataLab configuration file"""
    with execenv.context(unattended=True):
        h5files = [get_test_fnames("*.h5")[1]]
        execenv.print("Testing DataLab configuration settings:")
        for index, conf in enumerate(CONFIGS):
            name = f"CONFIG{index}"
            apply_conf(conf, name)
            with cdl_app_context(exec_loop=True) as qapp:
                win = app.create(splash=False, h5files=h5files)
                qapp.processEvents()
                check_conf(conf, name, win, h5files)
            h5files = None
        execenv.print("=> Everything is OK")


if __name__ == "__main__":
    test_config()



# Copyright (c) DataLab Platform Developers, BSD 3-Clause license, see LICENSE file.

"""
Image FFT application test.
"""

# pylint: disable=invalid-name  # Allows short reference names like x, y, ...
# guitest: show

from cdl.obj import ImageTypes, create_image_from_param, new_image_param
from cdl.tests import cdltest_app_context


def test_fft2d_app():
    """FFT application test."""
    with cdltest_app_context() as win:
        panel = win.imagepanel
        newparam = new_image_param(itype=ImageTypes.GAUSS, width=100, height=100)
        i1 = create_image_from_param(newparam)
        panel.add_object(i1)
        panel.processor.compute_fft()
        panel.processor.compute_ifft()


if __name__ == "__main__":
    test_fft2d_app()
